@@ -12,6 +12,7 @@
 #include "termpriv.h"
 #include "winimg.h"
 #include "sixel.h"
+#include "regis.h"
 
 // tempfile_t manipulation
 
@@ -255,16 +256,17 @@ winimg_new(imglist **ppimg, char * id, unsigned char * pixels, uint len,
   //printf("winimg alloc %d -> [%d]\n", (int)sizeof(imglist), img->imgi);
   if (!img)
     return false;
-#ifdef debug_img_list
-  printf("winimg_new [%d]->%p l %d t %d w %d h %d\n", img->imgi, pixels, left, top, width, height);
-#endif
 
   static int _imgi = 0;
 #ifdef debug_img_disp
-  printf("winimg_new %d @%d:%d\n", _imgi, left, top);
+  printf("winimg_new %d @%d:%d\n", _imgi, left, scrtop);
 #endif
   img->imgi = ++_imgi;
 
+#ifdef debug_img_list
+  printf("winimg_new [%d]->%p l %d t %d w %d h %d\n", img->imgi, pixels, left, scrtop, width, height);
+#endif
+
   img->pixels = pixels;
   img->hdc = NULL;
   img->hbmp = NULL;
@@ -282,7 +284,12 @@ winimg_new(imglist **ppimg, char * id, unsigned char * pixels, uint len,
   img->y = maxval(short);
 
   img->len = len;
-  if (len) {  // image format, not sixel
+  if ((int)len < 0) {  // ReGIS graphics
+    img->id = 0;
+    img->cwidth = cell_width;
+    img->cheight = cell_height;
+  }
+  else if (len) {  // Image (iTerm2 protocol)
     img->id = id ? strdup(id) : 0;
     img->cwidth = cell_width;
     img->cheight = cell_height;
@@ -389,7 +396,7 @@ winimg_new(imglist **ppimg, char * id, unsigned char * pixels, uint len,
     (void)crop_x, (void)crop_y, (void)crop_width, (void)crop_height;
 #endif
   }
-  else
+  else  // Sixel graphics
     img->id = 0;
 
   *ppimg = img;
@@ -646,6 +653,19 @@ draw_img(HDC dc, imglist * img)
 #endif
 }
 
+
+// callback function to flush incremental graphic output, used for ReGIS
+// - we could easier define this as an inner function but then 
+//   gcc would generate a trampoline for it which is undesirable
+// so we need to go the ugly way to shuffle parameters to global variables
+static HDC _dc, _rdc;
+static int _xlft, _grtop, _padwidth, _padheight;
+static void flush(void)
+{
+  StretchBlt(_dc, _xlft, _grtop, _padwidth, _padheight,
+             _rdc, 0, 0, _padwidth, _padheight, SRCCOPY);
+}
+
 void
 winimgs_paint(void)
 {
@@ -688,7 +708,7 @@ static bool previously_selected = false;
   img = backward_img_traversal ? term.imgs.last : term.imgs.first;
   imglist * next;
 #ifdef debug_img_over
-  printf("--------------- imglist loop\n");
+  printf("--------------- imglist loop vlines %lld disptop %d\n", term.virtuallines, term.disptop);
 #endif
   for (; img; img = next) {
     next = backward_img_traversal ? img->prev : img->next;
@@ -711,16 +731,20 @@ static bool previously_selected = false;
       // if the image is out of scrollback, collect it
       destrimg = img;
 #ifdef debug_img_list
-      printf("paint: destroy @%d h %d virt %lld sb %d\n", img->top, img->height, term.virtuallines, term.sblines);
+      printf("paint: destroy @%lld h %d virt %lld sb %d\n", img->top, img->height, term.virtuallines, term.sblines);
 #endif
 #ifdef debug_img_over
-      printf("@%d:%d destroy out [%d]\n", img->top - term.virtuallines - term.disptop, img->left, img->imgi);
+      printf("@%lld:%d destroy out [%d]\n", img->top - term.virtuallines - term.disptop, img->left, img->imgi);
 #endif
     } else {
       int left = img->left;
       // do not adjust horizontal scrolling here
 
       int top = img->top - term.virtuallines - term.disptop;
+      // todo:
+      // should we make some adjustment here to tame the effect of 
+      // line-wrapping on images?
+      // see disabled setting of term.virtuallines in term_reflow()
 
       // suppress repetitive image painting
       if (left == img->x && top == img->y && !force_imgs)
@@ -733,7 +757,7 @@ static bool previously_selected = false;
         // if the image is scrolled out, serialize it into a temp file
 #ifdef debug_img_list
         if (img->hdc)
-          printf("paint: hibernate img [%d] v@%d s@%d\n", img->imgi, img->top, top);
+          printf("paint: hibernate img [%d] v@%lld s@%d\n", img->imgi, img->top, top);
 #endif
         winimg_hibernate(img);
 #ifdef debug_img_over
@@ -741,7 +765,7 @@ static bool previously_selected = false;
 #endif
       } else {
 #ifdef debug_img_list
-        printf("paint: check img [%d] v@%d s@%d\n", img->imgi, img->top, top);
+        printf("paint: check img [%d] v@%lld s@%d\n", img->imgi, img->top, top);
 #endif
         // create img DC handle if not initialized, or resume from hibernate
         winimg_lazyinit(img);
@@ -793,12 +817,14 @@ static bool previously_selected = false;
 
 #ifdef scale_graphics_in_double_width_lines
 #warning no working implementation yet; not done in xterm either
-        termline * line = fetch_line(top);
-        ushort lattr = line->lattr;
-        release_line(line);
-        if ((lattr & LATTR_MODE) != LATTR_NORM) {
-          // fix position in double-width line:
-          // adjust below: left, xlft, ...width, xrgt, iwidth, iheight
+        if (top < term.rows) {
+          termline * line = fetch_line(top);
+          ushort lattr = line->lattr;
+          release_line(line);
+          if ((lattr & LATTR_MODE) != LATTR_NORM) {
+            // fix position in double-width line:
+            // adjust below: left, xlft, ...width, xrgt, iwidth, iheight
+          }
         }
 #endif
 
@@ -835,6 +861,15 @@ static bool previously_selected = false;
             iheight = img->cheight * cell_height * img->height / img->pixelheight;
           }
           int ibot = max(0, top * cell_height + iheight) + OFFSET + PADDING;
+
+          // pre-clear background of graphics area
+          // - for ReGIS graphics, this could be skipped in order to 
+          //   support iterative combination of subsequent ReGIS output;
+          //   however, this fails due to screen refresh and scrolling, 
+          //   so to really support that ReGIS feature, 
+          //   considerable effort would need to be put into a strategy 
+          //   to combine overlayed graphics on refresh
+
           // fill either background image or colour
           if (*cfg.background) {
             fill_background(dc, &(RECT){xlft + iwidth, ytop, xrgt, ibot});
@@ -848,6 +883,7 @@ static bool previously_selected = false;
             FillRect(dc, &(RECT){xlft, ibot, xrgt, ybot}, br);
             DeleteObject(br);
           }
+
           if (!img->len) {
             // sixel needs this in addition
             ExcludeClipRect(dc, xlft + iwidth, ytop, xrgt, ibot);
@@ -860,7 +896,128 @@ static bool previously_selected = false;
 #ifdef debug_img_list
           printf("paint: display img\n");
 #endif
-          if (img->len) {
+          if (img->len < 0) {  // Draw ReGIS graphics
+            // extract ReGIS DCS parameter
+            int regarg = -1 - img->len;
+
+            // width of the ReGIS background pad,
+            // rounded up to cell size and scaled with a zoomed window
+            int padwidth = cell_width * img->width;
+            int padheight = cell_height * img->height;
+            // anchor position of graphics
+            int grtop = top * cell_height + OFFSET + PADDING;
+
+            // determine zoom factor
+#define scale_regis true
+            float scale;
+            if (scale_regis)
+              //scale = (float)cell_width / (float)img->cwidth;
+              scale = min(padwidth / 800.0, padheight / 480.0);
+            else
+              scale = 1.0;
+
+            // size of the ReGIS graphics area
+            int rwidth = img->pixelwidth * scale;
+            // snap to 5% original width grid
+            rwidth = rwidth / 40 * 40;
+            // scale height to aspect ratio
+            int rheight = (rwidth - 1) * 480 / 800 + 1;
+
+            // fit size into pad
+            if (rheight > padheight) {
+              rheight = padheight;
+              // snap to 5% original height grid
+              rheight = rheight / 24 * 24;
+              // scale width to aspect ratio
+              rwidth = (rheight - 1) * 800 / 480 + 1;
+            }
+            //printf("--------------- img s %d×%d c %d×%d pix %d×%d (pad %d×%d) reg %d×%d\n", img->width, img->height, img->cwidth, img->cheight, img->pixelwidth, img->pixelheight, padwidth, padheight, rwidth, rheight);
+
+#define regis_use_separate_dc
+
+#ifdef regis_use_separate_dc
+            // set up ReGIS drawing area and background pad;
+            // arrange double buffering to avoid flickering
+            // - well, actually it turns out that this would not be needed,
+            // but we stay with the separate rendering DC anyway, see below
+            HDC rdc = CreateCompatibleDC(dc);
+            // set up drawing pad
+            HBITMAP hbm = CreateCompatibleBitmap(dc, padwidth, padheight);
+            (void)SelectObject(rdc, hbm);
+#else
+#warning transform placement fails on GDI output (incl. ExtTextOut)
+#warning so text will not be placed properly
+            // rendering directly to a common DC, 
+            // using handle name rdc for clarity;
+            // however, this does not seem to work with ReGIS output:
+            // we need to adjust the coordinate system, and while 
+            // this works nicely for GDI+ rendering, it is apparently 
+            // ignored for GDI rendering (???), 
+            // so ReGIS text would be displaced from the drawing area
+            HDC rdc = dc;
+            XFORM old_xform1;
+            bool coord_transformed1 = SetGraphicsMode(rdc, GM_ADVANCED);
+            if (coord_transformed1 && GetWorldTransform(rdc, &old_xform1)) {
+              XFORM xform = (XFORM){1.0, 0.0, 0.0, 1.0, xlft, grtop};
+              coord_transformed1 = ModifyWorldTransform(rdc, &xform, MWT_LEFTMULTIPLY);
+            }
+#endif
+
+#ifdef debug_regis
+            // mark terminal drawing pad 
+            // (drawing area rounded up to cell-size)
+            colour bg = RGB(80, 90, 100);  // show pad filling
+            HBRUSH br = CreateSolidBrush(bg);
+            FillRect(rdc, &(RECT){0, 0, padwidth, padheight}, br);
+            DeleteObject(br);
+#endif
+
+#ifdef zoom_transform
+#warning transform zooming fails on GDI output (incl. ExtTextOut)
+#warning so text will not be scaled properly
+#warning also, this used to spoil dashed-style lines at larger zoom factor
+            int coord_transformed2 = SetGraphicsMode(rdc, GM_ADVANCED);
+            XFORM old_xform2;
+            if (coord_transformed2 && GetWorldTransform(rdc, &old_xform2)) {
+              XFORM xform = (XFORM){scale, 0.0, 0.0, scale, 0, 0};
+              coord_transformed2 = ModifyWorldTransform(rdc, &xform, MWT_LEFTMULTIPLY);
+              scale = 1.0;
+            }
+#endif
+
+            // provide graphics flushing function with parameters;
+            // we could nicely define the flush function here locally 
+            // but then gcc would generate a trampoline which is undesirable
+            // so we need to shuffle parameters to global variables
+            _dc = dc; _rdc = rdc;
+            _xlft = xlft; _grtop = grtop;
+            _padwidth = padwidth; _padheight = padheight;
+            //printf("pad %d/%d draw %d/%d\n", padwidth, padheight, rwidth, rheight);
+
+            // scale while drawing
+            regis_draw(rdc, scale, rwidth, rheight, regarg, img->pixels, flush);
+
+#ifdef zoom_transform
+            if (coord_transformed2)
+              SetWorldTransform(rdc, &old_xform2);
+#endif
+
+#ifdef regis_use_separate_dc
+            // copy rendered graphics from the separate DC into the window DC
+            StretchBlt(dc, xlft, grtop, padwidth, padheight,
+                       rdc, 0, 0, padwidth, padheight, SRCCOPY);
+            DeleteObject(hbm);
+            DeleteDC(rdc);
+#else
+            // restore coordinate mapping of window DC
+            if (coord_transformed1)
+              SetWorldTransform(rdc, &old_xform1);
+#endif
+
+            ExcludeClipRect(dc, xlft, ytop, xrgt, ybot);
+          }
+          else if (img->len) {  // Draw Image (iTerm2 protocol)
+
             //draw_img(dc, img);
             // underlay image with background;
             // do it in a copy to avoid flickering
@@ -883,7 +1040,7 @@ static bool previously_selected = false;
             DeleteDC(hdc);
             ExcludeClipRect(dc, xlft, ytop, xrgt, ybot);
           }
-          else {
+          else {  // Draw Sixel graphis
             // adjust horizontal scrolling
             left -= horclip() / cell_width;
 
@@ -912,7 +1069,7 @@ static bool previously_selected = false;
           //destroy and remove
           destrimg = img;
 #ifdef debug_img_list
-          printf("paint: destroy @%d h %d virt %lld sb %d\n", img->top, img->height, term.virtuallines, term.sblines);
+          printf("paint: do destroy @%lld h %d virt %lld sb %d\n", img->top, img->height, term.virtuallines, term.sblines);
 #endif
 #ifdef debug_img_over
           printf("@%d:%d destroy over [%d]\n", top, left, img->imgi);
